# Copyright Commonwealth of Australia, Bureau of Meteorology 2024.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ruff: noqa: F401

"""
`pyearthtools.zoo`

Contains the PyEarthTools model registry and provides the command-line interaction
for easy inference with bundled models.

Each model can setup data and pipeline configurations to allow a user to choose the data location,
and use live data.

# Registration
    There are two ways to register a model for use with `pyearthtools.zoo`.

    One allows it to be dynamically imported, and the other only visible when the other package
    is manually imported.

## Entrypoints - Dynamic
    This method allows use of the model once its installed with no other action needed,

    Assign an entry point underneath `pyearthtools.zoo.model`.
    If this model needs to be categorised, seperate the category with `_`.

    Example: - inside the `pyproject.toml`.
    ```toml
        [project.entry-points."pyearthtools.zoo.model"]
        NESM_Model = "MODULE_NAME.registered_model:ModelClass"
    ```
    This will be available underneath `pyearthtools.zoo.Models.NESM.Model`, and when using the command line,
    at `pyearthtools-models predict 'NESM/Model' ...`

## Manual Registration
    This method will only expose the model once it has been manually imported.
    It takes use of decorators to register the class.

    If this model needs to be categorised, seperate the category with `/`.

    Example:
    ```python
    import pyearthtools.zoo

    @pyearthtools.zoo.register('Category/Name`)
    class MODEL():
        ....
    ```

    Once the model is imported, it will be accessible at
    `pyearthtools.zoo.Models.Category.Name` or `pyearthtools.zoo.Models['Category/Name']`

    The model will also be visible when using the `pyearthtools.zoo.predict/interactive/data` functions.

## Dynamic Importing
    Setting $pyearthtools_MODELS_IMPORTS can allow modules not in the environment to be imported when running `pyearthtools-models`
    to allow the commands to access custom models.

    Seperate modules by ':', and within each specification, split the name and path by '@'.

        i.e.

        ```shell
        export pyearthtools_MODELS_IMPORTS=MODULE_NAME@PATH_TO_MODULE_GOES_HERE:
        ```
    Be aware that depending on what is being imported, this may drastically reduce command responsiveness.
"""

from pyearthtools.zoo import config as _
from pyearthtools.zoo import logger as __

from pyearthtools.zoo.register import register
from pyearthtools.zoo.model import BaseForecastModel

from pyearthtools.zoo.warnings import AccessorRegistrationWarning
from pyearthtools.zoo.exceptions import ModelException, ModelRegistrationException

from pyearthtools.zoo import utils

Models = utils.AvailableModels()

from pyearthtools.zoo.predict import predict, data, interactive  # pylint: disable=C0413  # noqa: E402
from pyearthtools.zoo.commands import commands  # pylint: disable=C0413 # noqa: E402

__version__ = "0.5.0"


def available_models() -> tuple[str, ...]:
    """Get available models"""
    return Models.available


LIVE_SUBSTRINGS = ["live", "cds"]

__all__ = ["register", "BaseForecastModel", "Models", "predict", "data", "interactive"]



# Copyright Commonwealth of Australia, Bureau of Meteorology 2024.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=C0116,R0913,E1120,R1735

"""
`pyearthtools-models` commands

Available sub commands
```
- predict           - Predict
- interactive       - Predict Interactively
- models            - See what models are available
```
"""

from __future__ import annotations

import sys
from pathlib import Path
import logging

import click

import pyearthtools.zoo
from pyearthtools.zoo.commands import utils as command_utils
from pyearthtools.zoo.utils import AvailableModels

available_models = AvailableModels()

LOG = logging.getLogger("pyearthtools.zoo")


@click.group("pyearthtools-models")
@click.option("--debug", is_flag=True, help="Log debug")
@click.option("--info", is_flag=True, help="Log info")
def entry_point(debug, info):
    """
    `pyearthtools.zoo` commands
    """
    if debug or info:
        log_level = "DEBUG" if debug else "INFO"

        import pyearthtools.utils

        pyearthtools.utils.config.set({"logger.default.stream_logger_level": log_level})
        pyearthtools.utils.config.set({"logger.models.stream_logger_level": log_level})

        pyearthtools.utils.logger.initiate_logging("models")


@entry_point.command("models", help="Print available models.")
def models():
    from pyearthtools.zoo.register import dynamic_import

    dynamic_import()

    _models = pyearthtools.zoo.Models
    if len(_models.available) == 0:
        print("NO MODELS AVAILABLE.")
        sys.exit(0)

    print(_models)
    print("(Specify category with a '/' seperation.)")
    sys.exit(0)


HELP_STR = f"""
    Run Prediction of registered model

    \b
{available_models!s}

    * EACH MODEL MAY REQUIRE SPECIFIC KWARGS NOT HERE LISTED *

    All other keyword arguments will be passed to underlying model

    MODEL: Model to use

    """


@entry_point.command(
    "predict",
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    ),
    help=HELP_STR,
)
@click.pass_context
@click.argument("model", type=str)
@click.option("--time", type=str, required=True, help="Basetime to predict from")
@click.option("--pipeline_name", type=str, required=True, help="Pipeline config")
@click.option("--output", type=click.Path(file_okay=False), required=True, help="Output directory")
@click.option(
    "--data_cache",
    type=click.Path(file_okay=False),
    required=False,
    help="Data cache location",
)
@click.option(
    "--config_path",
    type=click.Path(file_okay=False),
    default=None,
    help="Override for config path",
)
def run_predict(
    ctx,
    model: str,
    time: str,
    output: str,
    pipeline_name: str,
    data_cache: str,
    config_path: str,
):
    ctx_kwargs = command_utils.get_keyword_from_ctx(ctx)

    if "data" in ctx_kwargs:
        raise ValueError("data has been deprecated as an argument for `predict`, use `data`.")

    predictions = pyearthtools.zoo.predict(
        model,
        time,
        pipeline_name=pipeline_name,
        output=output,
        data_cache=data_cache,
        config_path=config_path,
        **ctx_kwargs,
    )

    # FIXME: This is probably very much the wrong way to save data, but something
    # is needed as a stopgap while the train/predict/save cycle is restored
    # post migration, and the more appropriate use of the output directory
    # through pipelines will be revisited afterwards

    if output is not None:
        import os

        filename = "standard_filename.nc"
        filename = os.path.join(output, filename)
        predictions.to_netcdf(filename)

    print(f"Model Predictions saved underneath {output!r}.")


INTERACTIVE_HELP_STR = f"""
    Interactive Prediction

    Run Prediction of registered model, getting arguments interactively

    \b
{available_models!s}

    If a keyword argument is not given, this command will prompt the user,
    however, if passed, will skip.
    """


@entry_point.command(
    "interactive",
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    ),
    help=INTERACTIVE_HELP_STR,
)
@click.pass_context
@click.option("--model", type=str, required=False, help="Model to use", default=None)
@click.option("--time", type=str, required=False, help="Basetime to predict from", default=None)
@click.option("--pipeline_name", type=str, required=False, help="Pipeline config", default=None)
@click.option(
    "--output",
    type=click.Path(file_okay=False),
    required=False,
    help="Output directory",
    default=None,
)
@click.option(
    "--data_cache",
    type=click.Path(file_okay=False),
    required=False,
    help="Data cache location",
    default=None,
)
@click.option(
    "--config_path",
    type=click.Path(file_okay=False),
    default=None,
    help="Override for config path",
)
def interactive(
    ctx,
    model: str,
    time: str,
    pipeline_name: str,
    output: Path | str,
    data_cache: Path | str,
    config_path: str | Path,
):
    ctx_kwargs = command_utils.get_keyword_from_ctx(ctx)

    if "data" in ctx_kwargs:
        raise ValueError("data has been deprecated as an argument for `predict`, use `data`.")

    from pyearthtools.zoo.predict import _get_predict_kwargs_interactively

    predict_kwargs = _get_predict_kwargs_interactively(
        model=model,
        time=time,
        pipeline_name=pipeline_name,
        output=output,
        data_cache=data_cache,
        config_path=config_path,
        **ctx_kwargs,
    )

    kwargs_str = " ".join(
        f"--{key} {value!r}" for key, value in predict_kwargs.items() if value is not None and not key == "model"
    )

    print("Use this command to run non interactively:")
    print(f"\tpyearthtools-models predict {predict_kwargs['model']!r} {kwargs_str}".replace("  ", " "))
    pyearthtools.zoo.predict(quiet=predict_kwargs.pop("quiet", True), **predict_kwargs)

    print(f"Model Predictions saved underneath {predict_kwargs['output']!r}.")


DATA_HELP_STR = f"""
    Get data for model

    \b
{available_models!s}

    * EACH MODEL MAY REQUIRE SPECIFIC KWARGS NOT HERE LISTED *

    All other keyword arguments will be passed to underlying model

    MODEL: Model to use

    """


@entry_point.command(
    "data",
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    ),
    help=DATA_HELP_STR,
)
@click.pass_context
@click.argument("model", type=str)
@click.option("--time", type=str, required=True, help="Basetime to get data for")
@click.option("--pipeline_name", type=str, default=None, required=True, help="Pipeline config")
@click.option(
    "--data_cache",
    type=click.Path(file_okay=False),
    required=False,
    help="Data cache location",
)
@click.option(
    "--config_path",
    type=click.Path(file_okay=False),
    default=None,
    help="Override for config path",
)
def data(ctx, model, time, pipeline, data_cache, config_path: str | Path):
    ctx_kwargs = command_utils.get_keyword_from_ctx(ctx)
    _ = pyearthtools.zoo.data(model, time, pipeline, data_cache, config_path=config_path, **ctx_kwargs)

    print(f"All data was retrieved. {'Cached at ' + data_cache if data_cache is not None else ''}")
    print("Model ready for prediction.")


if __name__ == "__main__":
    entry_point()



# Copyright Commonwealth of Australia, Bureau of Meteorology 2024.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Command utility functions for `pyearthtools.zoo`
"""

from __future__ import annotations

from typing import Any

import click
from pyearthtools.zoo import utils


def parse_args_to_dict(*args: str) -> dict[str, Any]:
    """
    Convert a list of arguments into a dictionary.

    Of the format, [--key, value, --key2, value]

    Raises:
        KeyError:
            If cannot parse args

    Returns:
        (dict[str, Any]):
            Dictionary of args
    """
    d = {}

    def remove_surrounding(query: str, char: str = "'") -> str:
        """
        Remove prefix and suffix of char if present
        """
        return query.removeprefix(char).removesuffix(char)

    list_args = list(map(lambda x: remove_surrounding(x.strip()), args))

    while "" in list_args:
        list_args.remove("")

    if len(list_args) > 0:
        i = 0
        while i < len(list_args):
            if not str(list_args[i]).startswith("--"):
                raise KeyError(f"{args[i]} is an invalid keyword argument, ensure it starts with '--'")

            key = str(list_args[i]).replace("--", "")

            if i == (len(list_args) - 1) or (i < len(list_args) - 1 and list_args[i + 1].startswith("--")):
                d[key] = True
                i += 1
            else:
                d[key] = utils.parse_str(list_args[i + 1])
                i += 2
    return d


def get_keyword_from_ctx(ctx: click.Context) -> dict[str, Any]:
    """Convert click Context to a dictionary

    Args:
        ctx (click.Context):
            Extra click context

    Returns:
        (dict[str, Any]):
            Context as a dictionary
    """
    return parse_args_to_dict(*ctx.args)


def parse_str_to_dict(user_input: str) -> dict[str, Any]:
    """
    Parse user input string in form of click input to dictionary.

    >>> parse_str_to_dict('--variable value')
    {'variable': 'value'}

    Args:
        user_input (str):
            User input

    Returns:
        (dict[str, Any]):
            Dictionry of key:value
    """

    elements = user_input.replace("--", ";&;--").split(";&;")
    complete_split: list[str] = []
    for elem in elements:
        for sp in elem.split(" ", maxsplit=1):
            complete_split.append(sp)
    return parse_args_to_dict(*complete_split)



# Copyright Commonwealth of Australia, Bureau of Meteorology 2024.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any
import pytest

import pyearthtools.zoo


def list_to_category(
    arg: list[tuple[str, Any]],
) -> pyearthtools.zoo.utils.CategorisedObjects:
    cats = pyearthtools.zoo.utils.CategorisedObjects("Testing")
    arg.sort()

    for e in arg:
        if not isinstance(e, tuple):
            e = (e, True)  # Add true as value by default
        cats[e[0]] = e[1]
    return cats


@pytest.mark.parametrize(
    "category, result",
    [
        (["test"], ("test",)),
        (["test/model"], ("test/model",)),
        (["test/model", "test/model1"], ("test/model", "test/model1")),
        (["test/category1/model"], ("test/category1/model",)),
    ],
)
def test_available(category, result):
    categories = list_to_category(category)
    for r in result:
        assert r.split("/")[-1] in str(categories)
        assert r.split("/")[-1] in str(repr(categories))

    assert categories.available == result
    assert tuple(list(categories.keys())) == tuple(set(r.split("/")[0] for r in result))
    assert tuple(list(categories)) == tuple(set(r.split("/")[0] for r in result))
    # assert tuple(list(categories.values())) == tuple(r.split('/')[1] for r in result)


@pytest.mark.parametrize(
    "category, result",
    [
        (None, tuple()),
        ({"test": None}, ("test",)),
        ({"test": {"model": None}}, ("test/model",)),
        ({"test": {"model": None, "model_1": None}}, ("test/model", "test/model_1")),
        ({"test": {"category1": {"model": None}}}, ("test/category1/model",)),
    ],
)
def test_update(category, result):
    categories = pyearthtools.zoo.utils.CategorisedObjects("Testing")
    categories.update(category)
    assert categories.available == result


@pytest.mark.parametrize(
    "category, result",
    [
        (["test"], ("test",)),
        (["test/model"], ("test/model",)),
        (["test/model", "test/model1"], ("test/model", "test/model1")),
        (["test/category1/model"], ("test/category1/model",)),
    ],
)
def test_available__dir(category, result):
    categories = list_to_category(category)
    for r in result:
        assert r.split("/")[0] in categories.__dir__()
        assert r in categories.__dir__()


@pytest.mark.parametrize(
    "category, result",
    [
        ([("test", {"sub_cat": 1})], ("test/sub_cat",)),
    ],
)
def test_available1(category, result):
    categories = list_to_category(category)
    for r in result:
        assert r.split("/")[-1] in str(categories)
    assert categories.available == result


@pytest.mark.parametrize(
    "category, key, result",
    [
        ([("test", True)], "test", True),
        ([("test/wow", True)], "test/wow", True),
        ([("test/wow", True)], "wow", True),
    ],
)
def test_keys(category, key, result):
    categories = list_to_category(category)
    assert categories[key] == result


@pytest.mark.parametrize(
    "category, key, expectedError",
    [
        (["test"], "tes5t", IndexError),  # Wrong key
        (["test/model"], "test-model", IndexError),  # Invalid key
        (["test/model"], "test-model", IndexError),  # Invalid key
        (["test/model", "test1/model"], "model", ValueError),  # Not Unique
    ],
)
def test_getitem_fail(category, key, expectedError):
    with pytest.raises(expectedError):
        _obj = list_to_category(category)[key]


@pytest.mark.parametrize(
    "category, key, expectedError",
    [
        (["test"], "tes5t", AttributeError),  # Wrong key
        (["test/model"], "test-model", AttributeError),  # Invalid key
        (["test/model", "test1/model"], "model", ValueError),  # Not Unique
    ],
)
def test_getattr_fail(category, key, expectedError):
    with pytest.raises(expectedError):
        _obj = getattr(list_to_category(category), key)



"""Setup config"""

import os

import pyearthtools.utils
import yaml


def reconfigure():
    fn = os.path.join(os.path.dirname(__file__), "models.yaml")
    pyearthtools.utils.config.ensure_file(source=fn)

    with open(fn) as f:
        defaults = yaml.safe_load(f)

    pyearthtools.utils.config.update_defaults(defaults)


reconfigure()



# Copyright Commonwealth of Australia, Bureau of Meteorology 2024.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Model wrapper functions.

Allows easy usage of any model for prediction or data retrieval.
"""
# pylint: disable=R0914,R0915

from __future__ import annotations

from pathlib import Path
import logging
from typing import Any
import warnings

import pyearthtools.zoo
from pyearthtools.zoo.commands import utils as command_utils
from pyearthtools.zoo import utils, exceptions
from pyearthtools.zoo.register import dynamic_import

LOG = logging.getLogger("pyearthtools.zoo")


def _prefunctions():
    """Run functions before evaluating the prediction arguments"""

    dynamic_import()


def _initialise_model(  # pylint: disable=R0913
    model: str,
    pipeline_name: str,
    output: Path | str | None,
    data_cache: Path | str | None = None,
    config_path: Path | str | None = None,
    quiet: bool = False,
    **kwargs,
) -> pyearthtools.zoo.BaseForecastModel:
    """
    Get initialised model class from config

    Args:
        model (str):
            Model name to load
        pipeline_name (str):
            Used to determine the filename for loading the pipeline config
        output (Path | str):
            Location to save data
        data_cache (Path | str | None):
            Where to cache data. Defaults to None
        config_path (Path | str | None):
           Override for config path. Defaults to None
        quiet (bool):
            Whether to print documentation of model.
        kwargs (dict[Any, Any]):
            Extra keyword arguments to send to the model.

    Raises:
        ModelException:
            If `model` is unknown
        ValueError:
            If `pipeline` is unknown
        TypeError:
            If arguments for the `model` are missing

    Returns:
        pyearthtools.zoo.BaseForecastModel:
            Initalised model class.

    """

    LOG.info("Loading %s", model)
    _prefunctions()

    model_class = getattr(pyearthtools.zoo.Models, model, None)
    if model_class is None:
        raise exceptions.ModelException(f"`pyearthtools.zoo` cannot find {model!r}.\n{pyearthtools.zoo.Models}")

    doc = model_class.__doc__
    if doc and not quiet:
        print(doc)

    config_path = Path(config_path).resolve() if config_path else config_path

    # Get list of available pipelines from the config path (more than one may be present)
    valid_pipelines = list(model_class._valid_pipeline(config_path=config_path).keys())  # pylint: disable=W0212
    if not model_class.is_valid_pipeline(pipeline_name, config_path=config_path):
        raise ValueError(f"Cannot parse pipeline: {pipeline_name!r}. Valid: {valid_pipelines}")

    required, _ = utils.get_arguments(model_class)
    for done in set(["pipeline", "output", "data_cache", "kwargs", "config_path", *kwargs.keys()]).intersection(
        set(required.keys())
    ):
        required.pop(done)

    if any(map(lambda x: x in pipeline_name.lower(), pyearthtools.zoo.LIVE_SUBSTRINGS)) and data_cache is None:
        warnings.warn(
            "Downloaded data is specified but the `data_cache` has been left unset. This will use a default location.",
            RuntimeWarning,
        )

    if len(required.keys()) > 0:
        raise TypeError(
            f"Some required keyword arguments are missing, {list(required.keys())}.\nPlease specify them as keywords."
        )

    return model_class(
        pipeline_name,
        output=output,
        data_cache=data_cache,
        config_path=config_path,
        **kwargs,
    )


def data(
    model: str,
    time: str,
    pipeline: str,
    data_cache: Path | str | None = None,
    config_path: Path | str | None = None,
    **kwargs,
) -> list[Any]:
    """
    Get data needed for model to run,

    Can be used to precache data for 'live' runs.

    Args:
        model (str):
            Model name to load
        time (str):
            Isoformat of time to get data for
        pipeline (str):
            Pipeline config to use
        data_cache (Path | str | None):
            Where to cache data. Defaults to None
        config_path (Path | str | None):
            Override for config path. Defaults to None
        kwargs (dict[Any, Any]):
            Extra keyword arguments to send to the model.

    Raises:
        RuntimeError:
            If an error occured, catch it with nice error message.

    Returns:
        list[Any]:
            Loaded data needed for the model.
    """

    initialised_model = _initialise_model(model, pipeline, None, data_cache, config_path, **kwargs)

    try:
        return initialised_model.data(time)
    except Exception as e:
        raise RuntimeError(
            "Something went wrong, please raise an issue with this trace "
            "and the associated arguments back to `pyearthtools.zoo`"
        ) from e


def predict(  # pylint: disable=R0913
    model: str,
    time: str,
    pipeline_name: str,
    output: Path | str,
    data_cache: Path | str | None = None,
    config_path: Path | str | None = None,
    **kwargs,
) -> Any:
    """
    Run a prediction for a given model, pipeline, and time.

    Args:
        model (str):
            Model name to load
        time (str):
            Isoformat of time to run prediction for
        pipeline (str):
            Pipeline config to use
        output (Path | str):
            Location to save data
        data_cache (Path | str | None):
            Where to cache data. Defaults to None
        config_path (Path | str | None):
            Override for config path. Defaults to None
        kwargs (dict[Any, Any]):
            Extra keyword arguments to send to the model.

    Raises:
        RuntimeError:
            If an error occured, catch it with nice error message.

    Returns:
        (Any):
            Loaded Predictions.
    """
    import pyearthtools.data  # pylint: disable=W0621

    extra_kwargs_str = " ".join(f"--{key} {value!r}" for key, value in kwargs.items())
    data_cache_str = " " if data_cache is None else f" --data_cache {data_cache}"

    # Add a transform to add the command line to the history attributes
    post_transforms = pyearthtools.data.transforms.attributes.SetAttributes(
        history=f"pet predict {model!r} --pipeline_name {pipeline_name!r} --output {output!s}"
        f"{f' --config_path {config_path}' if config_path else ''} "
        f"--time {time}{data_cache_str} {extra_kwargs_str}".replace("  ", " ")
    ) + kwargs.pop("post_transforms", pyearthtools.data.TransformCollection())

    initialised_model = _initialise_model(
        model,
        pipeline_name,
        output,
        data_cache,
        config_path,
        post_transforms=post_transforms,
        **kwargs,
    )

    try:
        return initialised_model.run(time)
    except Exception as e:
        raise RuntimeError(
            "Something went wrong, please raise an issue with this trace "
            "and the associated arguments back to `pyearthtools.zoo`"
        ) from e


def _get_predict_kwargs_interactively(**kwargs: Any) -> dict[str, Any]:
    print("pyearthtools.zoo - Getting model kwargs interactively...\n")

    def prompt(
        key,
        prompt,
        values: list[str] | str | None = None,
        file: bool = False,
        show_options: bool = True,
    ) -> Any:
        if key in kwargs:
            if kwargs.get(key, None) is not None:
                return kwargs.pop(key)
            kwargs.pop(key, None)

        import readline

        readline.set_completer_delims("\t")
        readline.parse_and_bind("tab: complete")

        if file:
            readline.set_completer(utils.TabCompleter().path_completer)
        elif values is not None:
            readline.set_completer(utils.TabCompleter().create_list_completer(values))

        if values == []:
            values = None

        if isinstance(values, str):
            values = [values]

        parsed_result = None
        while parsed_result is None or parsed_result == "":
            parsed_result = utils.parse_str(
                input(
                    f"{prompt}?{(' ' + str(list(values))) if values is not None and show_options else ''},"
                    f"(Key: {key}): "
                )
            )
            if parsed_result is None or parsed_result == "":
                print("Must provide an entry...")

        return parsed_result

    _prefunctions()

    if kwargs.get("model", None) is None:
        print(pyearthtools.zoo.Models)
        print("(Specify model path with a '/' seperation.)")

    model = prompt(
        "model",
        "Which model would you like to use",
        list(pyearthtools.zoo.available_models()),
        show_options=False,
    )

    LOG.info("Loading %s", model)

    model_class = getattr(pyearthtools.zoo.Models, model)
    doc = model_class.__doc__  # pylint: disable=W0212

    if doc:
        print(doc)

    config_path = kwargs.pop("config_path", None)
    config_path = Path(config_path).resolve() if config_path else config_path

    valid_pipelines = list(model_class._valid_pipeline(config_path=config_path).keys())  # pylint: disable=W0212
    pipeline = prompt("pipeline", "Which pipeline / data source do you want to use", valid_pipelines)
    if not model_class.is_valid_pipeline(pipeline, config_path=config_path):
        raise ValueError(f"Cannot use pipeline: {pipeline}. Valid: {valid_pipelines}")

    data_cache = kwargs.pop("data_cache", None)
    if any(map(lambda x: x in pipeline.lower(), pyearthtools.zoo.LIVE_SUBSTRINGS)):
        data_cache = prompt("data_cache", "Where should 'downloaded' data be stored/found", file=True)

    output = prompt("output", "Where would you like to save the output (path)", file=True)

    required, noted_kwargs = utils.get_arguments(model_class)
    for done in set(["pipeline", "output", "data_cache", "kwargs", "config_path"]).intersection(set(required.keys())):
        required.pop(done)

    required_kwargs = {}
    for name, options in required.items():
        if name not in kwargs:
            print("Some required keyword arguments are missing...")
        required_kwargs[name] = prompt(name, f"{name!r}", options, file=False)

    remaining_kwargs = list(set(noted_kwargs.keys()).difference(kwargs.keys()))
    if "config_path" in remaining_kwargs:
        remaining_kwargs.remove("config_path")

    model_kwargs = input(f"Other kwargs {remaining_kwargs}? (Click Format): ")
    model_kwargs = command_utils.parse_str_to_dict(model_kwargs)

    time = prompt(
        "time",
        f"Basetime to {'predict' if not data else 'get data'} for",
        "isoformat",
        file=False,
    )

    extra_kwargs = dict(required_kwargs)
    extra_kwargs.update(model_kwargs)
    extra_kwargs.update(kwargs)

    return dict(  # pylint: disable=R1735
        model=model,
        pipeline=pipeline,
        output=output,
        config_path=str(config_path),
        time=time,
        data_cache=data_cache,
        **extra_kwargs,
    )  # pylint: disable=R1735


def interactive(
    **kwargs: str,
) -> Any:
    """
    Interactively run a prediction for a given model, pipeline, and time.

    Args:
        kwargs (dict[str, Any]):
            Specified kwargs to not prompt for.
            See `.predict` for a list of base ones.
    Raises:
        RuntimeError:
            If an error occured, catch it with nice error message.

    Returns:
        (Any):
            Loaded predictions.
    """
    predict_kwargs = _get_predict_kwargs_interactively(**kwargs)

    return predict(quiet=predict_kwargs.pop("quiet", True), **predict_kwargs)
